/*******************************************************************************
 * $Id: Islands.h,v 1.7 2004/03/01 02:40:08 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef _Islands_h_
#define _Islands_h_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gn/gnSequence.h"
#include "libMems/IntervalList.h"
#include "libMems/NumericMatrix.h"
#include "libMems/MatchList.h"

/**
 * A class to represent an island in an alignment.  Islands are generally
 * large insertions of a region of sequence relative to
 * another sequence.
 */
class Island{
public:
	int64 startI;
	int64 startJ;
	int64 lengthI;
	int64 lengthJ;
};

/**
 * Identifies gaps in the alignment between pairs of sequences that are longer than
 * some number of base pairs in length.  Prints islands to an output stream
 */
void simpleFindIslands( IntervalList& iv_list, uint island_size, ostream& island_out );
void findIslandsBetweenLCBs( IntervalList& iv_list, uint island_size, ostream& island_out );

/**
 *  Find regions in each sequence that do not belong to any LCB, add them to their own
 * Interval (LCB) in the IntervalList.
 */
void addUnalignedIntervals( IntervalList& iv_list );

/**
 * Identifies stretches of alignment existing in all sequences that doesn't
 * contain a gap larger than a particular size.  Such regions are considered
 * the backbone of the alignment.
 */
void simpleFindBackbone( IntervalList& iv_list, uint backbone_size, uint max_gap_size, vector< GappedAlignment >& backbone_regions );

/**
 * writes out a list of backbone regions
 */
void outputBackbone( const vector< GappedAlignment >& backbone_regions, ostream& backbone_out );

void allocateDetailList( IntervalList& iv_list, vector< pair< uint64, uint64 > >& detail_list );
void allocateDetailList( MatchList& iv_list, vector< pair< uint64, uint64 > >& detail_list );

void getLCBDetailList( Interval& iv, vector< pair< uint64, uint64 > >& detail_list );

void getLCBDetailList( MatchList& iv, vector< pair< uint64, uint64 > >& detail_list );

void PrintDetailList( uint seq_count, const vector< pair< uint64, uint64 > >& detail_list, ostream& os, boolean skip_zeros = true);
//void PrintDetailList( uint seq_count, vector< pair< uint64, uint64 > >& detail_list, ostream& os, boolean skip_zeros );

void TransformDistanceIdentity( NumericMatrix<double>& identity );

void DistanceMatrix( MatchList& mlist, NumericMatrix<double>& identity );

void IdentityMatrix( IntervalList& iv, NumericMatrix<double>& identity );
void IdentityMatrix( Interval& iv, NumericMatrix<double>& identity );
void DistanceMatrix( uint seq_count, const vector< pair< uint64, uint64 > >& detail_list, NumericMatrix<double>& distance );

#endif // _Islands_h_
